@@ -1,5 +1,5 @@
 /*
-% Copyright (C) 2003-2011 GraphicsMagick Group
+% Copyright (C) 2003-2013 GraphicsMagick Group
 % Copyright (C) 2002 ImageMagick Studio
 % Copyright 1991-1999 E. I. du Pont de Nemours and Company
 %
@@ -238,6 +238,11 @@ typedef enum
     XV_332_Format /* P7 332 */
   } PNMSubformat;
 
+#if defined(HAVE_OPENMP) && !defined(DisableSlowOpenMP)
+#  define PNMReadUseOpenMP 1
+#  define PNMReadThreads (Min(2,omp_get_max_threads()))
+#endif
+
 static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
 {
   char
@@ -276,9 +281,6 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
     max_value,
     samples_per_pixel;
 
-  Quantum
-    *scale;
-
   /*
     Open image file.
   */
@@ -563,7 +565,6 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
       number_pixels=image->columns*image->rows;
       if (number_pixels == 0)
         ThrowReaderException(CorruptImageError,NegativeOrZeroImageSize,image);
-      scale=(Quantum *) NULL;
       if (image->storage_class == PseudoClass)
         {
           /*
@@ -592,19 +593,6 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
                     }
             }
         }
-      if ((image->storage_class != PseudoClass) || (max_value > MaxRGB))
-        {
-          /*
-            Compute pixel scaling table.
-          */
-          scale=MagickAllocateMemory(Quantum *,
-                                     (max_value+1)*sizeof(Quantum));
-          if (scale == (Quantum *) NULL)
-            ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,
-                                 image);
-          for (i=0; i <= max_value; i++)
-            scale[i]=ScaleAnyToQuantum((unsigned long) i, max_value);
-        }
       if (image_info->ping && (image_info->subrange != 0))
         if (image->scene >= (image_info->subimage+image_info->subrange-1))
           break;
@@ -694,8 +682,6 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
 			ValidateScalingIndex(image, intensity, max_value);
 			if (EOFBlob(image))
 			  break;
-			if (scale != (Quantum *) NULL)
-			  intensity=scale[intensity];
 			index=intensity;
 			VerifyColormapIndex(image,index);
 			indexes[x]=index;
@@ -715,8 +701,7 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
 			ValidateScalingIndex(image, intensity, max_value);
 			if (EOFBlob(image))
 			  break;
-			if (scale != (Quantum *) NULL)
-			  intensity=scale[intensity];
+                        intensity=ScaleAnyToQuantum(intensity, max_value);
 			q->red=q->green=q->blue=intensity;
 			is_monochrome &= IsMonochrome(*q);
 			q++;
@@ -770,12 +755,9 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
                     if (EOFBlob(image))
                       break;
                     ValidateScalingPixel(image, pixel, max_value);
-                    if (scale != (Quantum *) NULL)
-                      {
-                        pixel.red=scale[pixel.red];
-                        pixel.green=scale[pixel.green];
-                        pixel.blue=scale[pixel.blue];
-                      }
+                    pixel.red=ScaleAnyToQuantum(pixel.red, max_value);
+                    pixel.green=ScaleAnyToQuantum(pixel.green, max_value);
+                    pixel.blue=ScaleAnyToQuantum(pixel.blue, max_value);
                     q->red=(Quantum) pixel.red;
                     q->green=(Quantum) pixel.green;
                     q->blue=(Quantum) pixel.blue;
@@ -822,7 +804,8 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
             MagickBool
 	      check_pixels,
               is_grayscale,
-              is_monochrome;
+              is_monochrome,
+              use_scaling;
         
             unsigned long
               row_count=0;
@@ -830,6 +813,17 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
             ThreadViewDataSet
               *scanline_set;
 
+            double
+              sample_scale;
+
+            unsigned int
+              sample_max;
+
+#if defined(HAVE_OPENMP) && !defined(DisableSlowOpenMP)
+            int
+              pnm_read_threads = PNMReadThreads;
+#endif
+
 	    (void) LogMagickEvent(CoderEvent,GetMagickModule(),"Reading PAM");
 
 	    ImportPixelAreaOptionsInit(&import_options);
@@ -843,6 +837,10 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
 	    */
 	    quantum_type=UndefinedQuantum;
 	    import_options.grayscale_miniswhite=MagickFalse;
+            sample_max=RoundDoubleToQuantum((MaxRGBDouble*max_value)/
+                                            MaxValueGivenBits(bits_per_sample));
+            sample_scale=MaxRGBDouble/sample_max;
+            use_scaling=(MaxRGB != sample_max);
 	    bytes_per_row=0;
 	    if (1 == samples_per_pixel)
 	      {
@@ -889,8 +887,6 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
 		  }
 	      }
 
-	    
-
 	    if (1 == samples_per_pixel)
 	      {
 		check_pixels=MagickFalse;
@@ -912,14 +908,12 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
             scanline_set=AllocateThreadViewDataArray(image,exception,bytes_per_row,1);
             if (scanline_set == (ThreadViewDataSet *) NULL)
               ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
-#if 1
-#if defined(HAVE_OPENMP) && !defined(DisableSlowOpenMP)
+#if PNMReadUseOpenMP
 #  if defined(TUNE_OPENMP)
 #    pragma omp parallel for schedule(runtime) shared(is_grayscale,is_monochrome,row_count,status)
 #  else
-#    pragma omp parallel for schedule(static,1) shared(is_grayscale,is_monochrome,row_count,status)
+#    pragma omp parallel for num_threads(pnm_read_threads) schedule(static,1) shared(is_grayscale,is_monochrome,row_count,status)
 #  endif
-#endif
 #endif
             for (y=0; y < (long) image->rows; y++)
               {
@@ -945,7 +939,7 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
 		ImportPixelAreaInfo
 		  import_info;
 
-#if defined(HAVE_OPENMP) && !defined(DisableSlowOpenMP)
+#if PNMReadUseOpenMP
 #  pragma omp critical (GM_ReadPNMImage)
 #endif
                 thread_status=status;
@@ -954,7 +948,7 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
             
                 pixels=AccessThreadViewData(scanline_set);
             
-#if defined(HAVE_OPENMP) && !defined(DisableSlowOpenMP)
+#if PNMReadUseOpenMP
 #  pragma omp critical (GM_ReadPNMImage)
 #endif
                 {
@@ -985,6 +979,26 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
                   if (!ImportImagePixelArea(image,quantum_type,bits_per_sample,pixels,
 					    &import_options,&import_info))
                     thread_status=MagickFail;
+                /*
+                  Scale sub-ranged pixels up to full range if necessary
+                */
+                if ((thread_status != MagickFail) && (use_scaling))
+                  for (x=0; x < image->columns; x++)
+                    {
+                      SetRedSample(&q[x],
+                                   RoundDoubleToQuantum(sample_scale*
+                                                        GetRedSample(&q[x])));
+                      SetGreenSample(&q[x],
+                                     RoundDoubleToQuantum(sample_scale*
+                                                          GetGreenSample(&q[x])));
+                      SetBlueSample(&q[x],
+                                    RoundDoubleToQuantum(sample_scale*
+                                                         GetBlueSample(&q[x])));
+                      if (image->matte)
+                        SetOpacitySample(&q[x],
+                                         RoundDoubleToQuantum(sample_scale*
+                                                              GetOpacitySample(&q[x])));
+                    }
                 /*
                   For a DirectClass image, check all pixels for
                   gray/monochrome status since this format is often
@@ -996,20 +1010,19 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
                 if (thread_status != MagickFail)
 		  if (check_pixels)
 		    if (thread_is_grayscale || thread_is_monochrome)
-		      for (x=image->columns; x; x--)
+		      for (x=0; x < image->columns; x++)
 			{
-			  thread_is_grayscale = thread_is_grayscale && IsGray(*q);
-			  thread_is_monochrome = thread_is_monochrome && IsMonochrome(*q);
+			  thread_is_grayscale = thread_is_grayscale && IsGray(q[x]);
+			  thread_is_monochrome = thread_is_monochrome && IsMonochrome(q[x]);
 			  if (!thread_is_grayscale && !thread_is_monochrome)
 			    break;
-			  q++;
 			}
 
                 if (thread_status != MagickFail)
                   if (!SyncImagePixels(image))
                     thread_status=MagickFail;
 
-#if defined(HAVE_OPENMP) && !defined(DisableSlowOpenMP)
+#if PNMReadUseOpenMP
 #  pragma omp critical (GM_ReadPNMImage)
 #endif
                 {
@@ -1034,8 +1047,6 @@ static Image *ReadPNMImage(const ImageInfo *image_info,ExceptionInfo *exception)
         default:
           ThrowReaderException(CorruptImageError,ImproperImageHeader,image);
         }
-      if (scale != (Quantum *) NULL)
-        MagickFreeMemory(scale);
       /*
         Proceed to next image.
       */
@@ -1298,7 +1309,7 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
   scene=0;
   do
     {
-      depth=(image->depth <= 8 ? 8 : 16);
+      depth=(image->depth <= 8 ? 8 : image->depth <= 16 ? 16 : 32);
 
       /*
 	Write PNM file header.
@@ -1528,11 +1539,9 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
 
 	    i=0;
 	    j=0;
-	    
-	    if (depth <= 8)
-	      value=255;
-	    else
-	      value=65535;
+
+            value=(depth <=8 ? 255 : depth <= 16 ? 65535 : 4294967295);
+
 	    j += sprintf(&buffer[j],"%u\n",value);
 	    for (y=0; y < image->rows; y++)
 	      {
@@ -1553,12 +1562,19 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
 			/* Use LUT for speed */
 			value=ScaleQuantumToChar(index);
 			AppendUnsignedCharValueToString(j,buffer,value);
+                        buffer[j++] = ' ';
 		      }
-		    else
+                    else if (depth <= 16)
 		      {
 			value=ScaleQuantumToShort(index);
 			j += sprintf(&buffer[j]," %u",value);
 		      }
+                    else
+                      {
+			value=ScaleQuantumToLong(index);
+			j += sprintf(&buffer[j]," %u",value);
+                      }
+
 		    i++;
 		    if (i == 12)
 		      {
@@ -1614,10 +1630,8 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
 	    i=0;
 	    j=0;
 
-	    if (depth <= 8)
-	      value=255;
-	    else
-	      value=65535;
+            value=(depth <=8 ? 255 : (depth <= 16 ? 65535 : 4294967295));
+
 	    j += sprintf(&buffer[j],"%u\n",value);
 	    for (y=0; y < image->rows; y++)
 	      {
@@ -1642,10 +1656,19 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
 			AppendUnsignedCharValueToString(j,buffer,value);
 			buffer[j++] = ' ';
 		      }
+                    else if (depth <= 16)
+                      {
+			j += sprintf(&buffer[j],"%u %u %u ",
+                                     ScaleQuantumToShort(p->red),
+				     ScaleQuantumToShort(p->green),
+                                     ScaleQuantumToShort(p->blue));
+                      }
 		    else
 		      {
-			j += sprintf(&buffer[j],"%u %u %u ",ScaleQuantumToShort(p->red),
-				     ScaleQuantumToShort(p->green),ScaleQuantumToShort(p->blue));
+			j += sprintf(&buffer[j],"%u %u %u ",
+                                     (unsigned int) ScaleQuantumToLong(p->red),
+				     (unsigned int) ScaleQuantumToLong(p->green),
+                                     (unsigned int) ScaleQuantumToLong(p->blue));
 		      }
 		    i++;
 		    if (i == 4)
@@ -1710,7 +1733,7 @@ static unsigned int WritePNMImage(const ImageInfo *image_info,Image *image)
 	    /*
 	      Deduce correct export parameters.
 	    */
-	    bits_per_sample=(depth > 8 ? 16 : 8);
+            bits_per_sample=(depth <=8 ? 8 : (depth <= 16 ? 16 : 32));
 	    quantum_type=RGBQuantum;
 	    if (PBM_RAW_Format == format)
 	      {
