@@ -1,6 +1,5 @@
 /*
-% Copyright (C) 2003, 2005 GraphicsMagick Group
-% Copyright (C) 2000-2002, Ghostgum Software Pty Ltd.  All rights reserved.
+% Copyright (C) 2003 - 2010 GraphicsMagick Group
 % Copyright (C) 2002 ImageMagick Studio
 %
 % This program is covered by multiple licenses, which are described in
@@ -909,12 +908,6 @@ NTGhostscriptFind(const char **gs_productfamily,
 	PHKEY phkResult)
       */
       open_key_mode=KEY_READ;
-#if defined(KEY_WOW64_32KEY)
-      // Access a 32-bit key from either a 32-bit or 64-bit
-      // application.  This flag is not supported on Windows 2000.
-      // Presumably Ghostscript is registered in the 32-bit registry.
-      open_key_mode |= KEY_WOW64_32KEY;
-#endif // defined(KEY_WOW64_32KEY)
       if ((winstatus=RegOpenKeyExA(hkeyroot, key, 0, open_key_mode, &hkey))
 	  == ERROR_SUCCESS)
 	{
@@ -1064,7 +1057,7 @@ NTGhostscriptGetString(const char *name, char *ptr, const size_t len)
   
   for (i=0; i < sizeof(hkeys)/sizeof(hkeys[0]); ++i)
     {
-      length = len;
+      length = (int) len;
       if (NTGetRegistryValue(hkeys[i].hkey, key, name, ptr, &length) == 0)
 	{
 	  (void) LogMagickEvent(ConfigureEvent,GetMagickModule(),
@@ -1173,7 +1166,7 @@ MagickExport const GhostscriptVectors *NTGhostscriptDLLVectors( void )
 %   Method NTGhostscriptEXE obtains the path to the latest Ghostscript
 %   executable.  The method returns TRUE if path is updated, otherwise
 %   FALSE.  When the full path value is not obtained, then the value
-%   "gswin32c.exe" is used.
+%   "gswin32c.exe" or "gswin64c.exe" is used.
 %
 %  The format of the NTGhostscriptEXE method is:
 %
@@ -1192,13 +1185,17 @@ MagickExport int NTGhostscriptEXE(char *path, int path_length)
     cache[MaxTextExtent],
     *result=NULL;
 
-  char
-    *p;
-
   if (NULL == result)
     {
+      char
+        gsexe[MaxTextExtent];
+
+      char
+        *p;
+
       /* Ensure a suitable default. */
-      (void) strlcpy(cache,"gswin32c.exe",sizeof(cache));
+      (void) FormatString(gsexe,"gswin%uc.exe",(unsigned int) sizeof(p)*8);
+      (void) strlcpy(cache,gsexe,sizeof(cache));
 
       if (NTGhostscriptDLL(cache,sizeof(cache)))
 	{
@@ -1206,7 +1203,7 @@ MagickExport int NTGhostscriptEXE(char *path, int path_length)
 	  if (p) {
 	    p++;
 	    *p = 0;
- 	    (void) strlcat(cache,"gswin32c.exe",sizeof(cache));
+            (void) strlcat(cache,gsexe,sizeof(cache));
 	  }
 	}
       result=cache;
@@ -1258,8 +1255,17 @@ MagickExport int NTGhostscriptFonts(char *path, int path_length)
 
   /*
     The format of the GS_LIB string is a path similar to
-    "c:\gs8.14\lib;C:\gs\fonts;C:\gs\gs8.14\Resource". Ghostscript
-    7.0X does not include the Resource entry.
+
+    "C:\Program Files\gs\gs9.10\bin;C:\Program Files\gs\gs9.10\lib;C:\Program Files\gs\gs9.10\fonts"
+
+    Ghostscript 7.0X does not include the Resource entry.
+
+    Search path used by GPL Ghostscript 9.10 (2013-08-30):
+      C:\Program Files\gs\gs9.10\bin ; C:\Program Files\gs\gs9.10\lib ;
+      C:\Program Files\gs\gs9.10\fonts ; %rom%Resource/Init/ ; %rom%lib/ ;
+      c:/gs/gs9.10/Resource/Init ; c:/gs/gs9.10/lib ;
+      c:/gs/gs9.10/Resource/Font ; c:/gs/fonts
+
   */
   {
     const char
@@ -1274,14 +1280,14 @@ MagickExport int NTGhostscriptFonts(char *path, int path_length)
           font_dir_file[MaxTextExtent];
             
         const char
-          *seperator;
+          *separator;
             
         int
           length;
             
-        seperator = strchr(start,DirectoryListSeparator);
-        if (seperator)
-          length=seperator-start;
+        separator = strchr(start,DirectoryListSeparator);
+        if (separator)
+          length=separator-start;
         else
           length=end-start;
         (void) strlcpy(font_dir,start,Min(length+1,MaxTextExtent));
@@ -1299,6 +1305,36 @@ MagickExport int NTGhostscriptFonts(char *path, int path_length)
         start += length+1;
       }
   }
+  {
+    /*
+      Check "c:/gs/fonts" since Ghostscript also looks there.  This
+      may be a more convenient place to put fonts since it would be
+      used by every Ghostscript installation on the system.
+
+      This part of the path to check is hard-coded in the Ghostscript
+      binary via AROOTDIR=c:/gs in base/msvclib.mak and it is highly
+      unlikely that Windows users will use a Ghostscript built with a
+      different GS_LIB_DEFAULT (which includes AROOTDIR) definition.
+    */
+
+    const char *
+      gs_font_dir           = "c:\\gs\\fonts";
+
+    char
+      font_dir_file[MaxTextExtent];
+
+    (void) strlcpy(font_dir_file,gs_font_dir,MaxTextExtent);
+    (void) strlcat(font_dir_file,DirectorySeparator,MaxTextExtent);
+    (void) strlcat(font_dir_file,"fonts.dir",MaxTextExtent);
+    if (IsAccessible(font_dir_file))
+      {
+        (void) strlcpy(path,gs_font_dir,path_length);
+        (void) LogMagickEvent(ConfigureEvent,GetMagickModule(),
+                              "Ghostscript fonts in directory \"%s\"",
+                              path);
+        return TRUE;
+      }
+  }
 
   return FALSE;
 }
@@ -2195,7 +2231,7 @@ MagickExport struct dirent *NTreaddir(DIR *entry)
       sizeof(entry->file_info.d_name))
     return ((struct dirent *) NULL);
   entry->firsttime=FALSE;
-  entry->file_info.d_namlen=strlen(entry->file_info.d_name);
+  entry->file_info.d_namlen=(int) strlen(entry->file_info.d_name);
   return (&entry->file_info);
 }
 
