@@ -138,7 +138,11 @@ MagickExport Image *ChopImage(const Image *image,const RectangleInfo *chop_info,
     Extract chop image.
   */
 #if defined(HAVE_OPENMP)  && !defined(DisableSlowOpenMP)
-#  pragma omp parallel for schedule(static,4) shared(row_count, status)
+#  if defined(TUNE_OPENMP)
+#    pragma omp parallel for schedule(runtime) shared(row_count, status)
+#  else
+#    pragma omp parallel for schedule(static,4) shared(row_count, status)
+#  endif
 #endif
   for (y=0; y < (long) clone_info.y; y++)
     {
@@ -209,7 +213,11 @@ MagickExport Image *ChopImage(const Image *image,const RectangleInfo *chop_info,
     Extract chop image.
   */
 #if defined(HAVE_OPENMP) && !defined(DisableSlowOpenMP)
-#  pragma omp parallel for schedule(static,4) shared(row_count, status)
+#  if defined(TUNE_OPENMP)
+#    pragma omp parallel for schedule(runtime) shared(row_count, status)
+#  else
+#    pragma omp parallel for schedule(static,4) shared(row_count, status)
+#  endif
 #endif
   for (y=0; y < (long) (image->rows-(clone_info.y+clone_info.height)); y++)
     {
@@ -529,7 +537,11 @@ MagickExport Image *CropImage(const Image *image,const RectangleInfo *geometry,
   if ((geometry->width == 0) || (geometry->height == 0))
     (void) memset(&crop_image->page,0,sizeof(RectangleInfo));
 #if defined(HAVE_OPENMP) && !defined(DisableSlowOpenMP)
-#  pragma omp parallel for schedule(static,4) shared(row_count, status)
+#  if defined(TUNE_OPENMP)
+#    pragma omp parallel for schedule(runtime) shared(row_count, status)
+#  else
+#    pragma omp parallel for schedule(static,4) shared(row_count, status)
+#  endif
 #endif
   for (y=0; y < (long) crop_image->rows; y++)
     {
@@ -803,7 +815,7 @@ MagickExport Image *DeconstructImages(const Image *image,
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %
 %  Use ExtentImage() to change the image dimensions as specified by geometry
-%  width and hight.  The existing image content is composited at the position
+%  width and height.  The existing image content is composited at the position
 %  specified by geometry x and y using the image compose method.  Existing
 %  image content which falls outside the bounds of the new image dimensions
 %  is discarded.
@@ -992,7 +1004,11 @@ MagickExport Image *FlipImage(const Image *image,ExceptionInfo *exception)
     Flip each row.
   */
 #if defined(HAVE_OPENMP) && !defined(DisableSlowOpenMP)
-#  pragma omp parallel for schedule(static,4) shared(row_count, status)
+#  if defined(TUNE_OPENMP)
+#    pragma omp parallel for schedule(runtime) shared(row_count, status)
+#  else
+#    pragma omp parallel for schedule(static,4) shared(row_count, status)
+#  endif
 #endif
   for (y=0; y < (long) flip_image->rows; y++)
     {
@@ -1114,7 +1130,11 @@ MagickExport Image *FlopImage(const Image *image,ExceptionInfo *exception)
     Flop each row.
   */
 #if defined(HAVE_OPENMP) && !defined(DisableSlowOpenMP)
-#  pragma omp parallel for schedule(static,4) shared(row_count, status)
+#  if defined(TUNE_OPENMP)
+#    pragma omp parallel for schedule(runtime) shared(row_count, status)
+#  else
+#    pragma omp parallel for schedule(static,4) shared(row_count, status)
+#  endif
 #endif
   for (y=0; y < (long) flop_image->rows; y++)
     {
@@ -1253,12 +1273,16 @@ MagickExport Image *MosaicImages(const Image *image,ExceptionInfo *exception)
   {
     page.x=next->page.x;
     page.y=next->page.y;
-    if ((next->columns+page.x) > page.width)
-      page.width=next->columns+page.x;
+    /*
+      Without casts, unsigned underflow can occur here if page offset
+      is negative and has greater magnitude than image size.
+    */
+    if (((long) next->columns+page.x) > (long) page.width)
+      page.width=(long) next->columns+page.x;
     if (next->page.width > page.width)
       page.width=next->page.width;
-    if ((next->rows+page.y) > page.height)
-      page.height=next->rows+page.y;
+    if (((long) next->rows+page.y) > (long) page.height)
+      page.height=(long) next->rows+page.y;
     if (next->page.height > page.height)
       page.height=next->page.height;
   }
